@@ -27,6 +27,7 @@
 #include "exm-error-dialog.h"
 #include "exm-installed-page.h"
 #include "exm-screenshot-view.h"
+#include "exm-types.h"
 #include "exm-upgrade-assistant.h"
 #include "local/exm-extension.h"
 #include "local/exm-manager.h"
@@ -67,6 +68,14 @@ enum {
 
 static GParamSpec *properties [N_PROPS];
 
+enum {
+    SIGNAL_0,
+    SIGNAL_SEARCH_CHANGED,
+    N_SIGNALS
+};
+
+static guint signals [N_SIGNALS];
+
 static void
 exm_window_get_property (GObject    *object,
                          guint       prop_id,
@@ -85,12 +94,6 @@ exm_window_get_property (GObject    *object,
     }
 }
 
-GtkSearchBar *
-exm_window_get_search_bar (ExmWindow *self)
-{
-  return self->search_bar;
-}
-
 static void
 extension_open_prefs (GtkWidget  *widget,
                       const char *action_name G_GNUC_UNUSED,
@@ -121,10 +124,8 @@ extension_remove_dialog_response (AdwAlertDialog   *dialog,
 {
     const char *response = adw_alert_dialog_choose_finish (dialog, result);
 
-    if (strcmp(response, "yes") == 0)
-    {
+    if (g_str_equal (response, "yes"))
         exm_manager_remove_extension (data->manager, data->extension);
-    }
 
     g_clear_pointer (&data->manager, g_object_unref);
     g_clear_pointer (&data->extension, g_object_unref);
@@ -153,25 +154,60 @@ extension_remove (GtkWidget  *widget,
                              (GAsyncReadyCallback) extension_remove_dialog_response, data);
 }
 
-static void
-on_install_done (GObject       *source,
-                 GAsyncResult  *res,
-                 gpointer       user_data G_GNUC_UNUSED)
-{
-    GError *error = NULL;
-    if (!exm_manager_install_finish (EXM_MANAGER (source), res, &error) && error)
-    {
-        // TODO: Properly log this
-        g_critical ("%s\n", error->message);
-    }
-}
-
 typedef struct
 {
     ExmManager *manager;
     gchar *uuid;
 } UnsupportedDialogData;
 
+static void
+on_install_done (GObject      *source,
+                 GAsyncResult *res,
+                 gpointer      user_data)
+{
+    ExmWindow *self = NULL;
+    UnsupportedDialogData *data = NULL;
+    ExmManager *manager = NULL;
+    ExmInstallButtonState state;
+    GError *error = NULL;
+
+    if (G_IS_OBJECT (user_data))
+    {
+        self = (ExmWindow *)user_data;
+        manager = self->manager;
+        state = EXM_INSTALL_BUTTON_STATE_DEFAULT;
+    }
+    else
+    {
+        data = (UnsupportedDialogData *)user_data;
+        manager = data->manager;
+        state = EXM_INSTALL_BUTTON_STATE_UNSUPPORTED;
+    }
+
+    if (!exm_manager_install_finish (source, res, &error))
+    {
+        // TODO: Properly log this
+        if (error)
+        {
+            g_critical ("%s\n", error->message);
+            g_clear_error (&error);
+        }
+
+        g_signal_emit_by_name (manager, "install-status", state);
+    }
+    else
+    {
+        g_signal_emit_by_name (manager, "install-status", EXM_INSTALL_BUTTON_STATE_INSTALLED);
+    }
+
+    if (data)
+    {
+        g_clear_pointer (&data->manager, g_object_unref);
+        g_clear_pointer (&data->uuid, g_free);
+        g_free (data);
+    }
+}
+
 static void
 extension_unsupported_dialog_response (AdwAlertDialog        *dialog,
                                        GAsyncResult          *result,
@@ -179,16 +215,20 @@ extension_unsupported_dialog_response (AdwAlertDialog        *dialog,
 {
     const char *response = adw_alert_dialog_choose_finish (dialog, result);
 
-    if (strcmp(response, "install") == 0)
+    if (g_str_equal (response, "install"))
     {
         exm_manager_install_async (data->manager, data->uuid, NULL,
                                    (GAsyncReadyCallback) on_install_done,
-                                   NULL);
+                                   data);
     }
+    else
+    {
+        g_signal_emit_by_name (data->manager, "install-status", EXM_INSTALL_BUTTON_STATE_UNSUPPORTED);
 
-    g_clear_pointer (&data->manager, g_object_unref);
-    g_clear_pointer (&data->uuid, g_free);
-    g_free (data);
+        g_clear_pointer (&data->manager, g_object_unref);
+        g_clear_pointer (&data->uuid, g_free);
+        g_free (data);
+    }
 }
 
 static void
@@ -208,6 +248,7 @@ extension_install (GtkWidget  *widget,
         UnsupportedDialogData *data = g_new0 (UnsupportedDialogData, 1);
         data->manager = g_object_ref (self->manager);
         data->uuid = g_strdup (uuid);
+        g_free (uuid);
 
         adw_alert_dialog_choose (self->unsupported_dialog, widget, NULL,
                                  (GAsyncReadyCallback) extension_unsupported_dialog_response, data);
@@ -318,15 +359,13 @@ search_online (GtkWidget  *widget,
                GVariant   *param G_GNUC_UNUSED)
 {
     ExmWindow *self;
-    GtkSearchEntry *search_entry;
     const char *search_text;
 
     self = EXM_WINDOW (widget);
 
-    search_entry = exm_browse_page_get_search_entry (self->browse_page);
     search_text = gtk_editable_get_text (GTK_EDITABLE (gtk_search_bar_get_child (self->search_bar)));
     adw_view_stack_set_visible_child_name (self->view_stack, "browse");
-    gtk_editable_set_text (GTK_EDITABLE (search_entry) , search_text);
+    exm_browse_page_search (self->browse_page, search_text);
     gtk_toggle_button_set_active (self->search_button, FALSE);
 }
 
@@ -365,26 +404,42 @@ on_visible_page_changed (AdwViewStack *view_stack,
                             is_installed_page);
 }
 
-void
-on_visible_installed_stack_changed (GObject    *object G_GNUC_UNUSED,
-                                    GParamSpec *pspec G_GNUC_UNUSED,
-                                    gpointer    user_data)
+const char *
+exm_window_get_search_query (ExmWindow *self)
 {
-    ExmWindow *self = (ExmWindow *) user_data;
-    gboolean search_mode;
-    GtkFilterListModel *search_list_model;
-    GtkStack *stack;
+    GtkSearchEntry *search_entry;
+    const char *query;
 
-    search_mode = gtk_search_bar_get_search_mode (self->search_bar);
-    search_list_model = exm_installed_page_get_search_list_model (self->installed_page);
-    stack = exm_installed_page_get_stack (self->installed_page);
+    search_entry = GTK_SEARCH_ENTRY (gtk_search_bar_get_child (self->search_bar));
+    query = gtk_editable_get_text (GTK_EDITABLE (search_entry));
 
-    if (search_mode && g_list_model_get_n_items (G_LIST_MODEL (search_list_model)) > 0)
-        gtk_stack_set_visible_child_name (stack , "page_results");
-    else if (search_mode)
-        gtk_stack_set_visible_child_name (stack , "page_empty");
-    else
-        gtk_stack_set_visible_child_name (stack , "page_list");
+    return query;
+}
+
+gboolean
+exm_window_get_search_mode (ExmWindow *self)
+{
+    return gtk_search_bar_get_search_mode (self->search_bar);
+}
+
+static void
+on_search_mode_enabled (GObject    *object G_GNUC_UNUSED,
+                        GParamSpec *pspec G_GNUC_UNUSED,
+                        gpointer    user_data)
+{
+    ExmWindow *self = (ExmWindow *)user_data;
+
+    exm_installed_page_show_page (gtk_search_bar_get_search_mode (self->search_bar),
+                                  self->installed_page);
+}
+
+static void
+on_search_changed (GtkSearchEntry *search_entry G_GNUC_UNUSED,
+                   gpointer        user_data)
+{
+    ExmWindow *self = (ExmWindow *)user_data;
+
+    g_signal_emit_by_name (self, "search-changed", NULL);
 }
 
 static gboolean
@@ -408,8 +463,7 @@ search_open_cb (GtkWidget *widget,
   }
   else if (g_strcmp0 (visible_stack_name, "browse") == 0)
   {
-      GtkSearchEntry *search_entry = exm_browse_page_get_search_entry (self->browse_page);
-      gtk_widget_grab_focus (GTK_WIDGET (search_entry));
+      exm_browse_page_focus_entry (self->browse_page);
   }
   else
   {
@@ -447,6 +501,14 @@ exm_window_class_init (ExmWindowClass *klass)
 
     g_object_class_install_properties (object_class, N_PROPS, properties);
 
+    signals [SIGNAL_SEARCH_CHANGED]
+        = g_signal_new ("search-changed",
+                        G_TYPE_FROM_CLASS (object_class),
+                        G_SIGNAL_RUN_LAST|G_SIGNAL_NO_RECURSE|G_SIGNAL_NO_HOOKS,
+                        0, NULL, NULL, NULL,
+                        G_TYPE_NONE, 0,
+                        NULL);
+
     GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
 
     gtk_widget_class_set_template_from_resource (widget_class, g_strdup_printf ("%s/exm-window.ui", RESOURCE_PATH));
@@ -467,7 +529,8 @@ exm_window_class_init (ExmWindowClass *klass)
     gtk_widget_class_bind_template_child (widget_class, ExmWindow, unsupported_dialog);
 
     gtk_widget_class_bind_template_callback (widget_class, on_visible_page_changed);
-    gtk_widget_class_bind_template_callback (widget_class, on_visible_installed_stack_changed);
+    gtk_widget_class_bind_template_callback (widget_class, on_search_mode_enabled);
+    gtk_widget_class_bind_template_callback (widget_class, on_search_changed);
 
     // TODO: Refactor ExmWindow into a separate ExmController and supply the
     // necessary actions/methods/etc in there. A reference to this new object can
@@ -505,9 +568,8 @@ exm_window_init (ExmWindow *self)
 
     gtk_widget_init_template (GTK_WIDGET (self));
 
-    if (strstr (APP_ID, ".Devel") != NULL) {
+    if (strstr (APP_ID, ".Devel") != NULL)
         gtk_widget_add_css_class (GTK_WIDGET (self), "devel");
-    }
 
     self->manager = exm_manager_new ();
     g_signal_connect (self->manager, "error-occurred", G_CALLBACK (on_error), self);
