@@ -21,7 +21,6 @@
 
 #include "exm-manager.h"
 
-#include "exm-extension.h"
 #include "shell-dbus-interface.h"
 
 #include "../exm-types.h"
@@ -56,6 +55,7 @@ enum {
     SIGNAL_0,
     SIGNAL_UPDATES_AVAILABLE,
     SIGNAL_ERROR_OCCURRED,
+    SIGNAL_INSTALL_STATUS,
     N_SIGNALS
 };
 
@@ -353,29 +353,6 @@ exm_manager_is_installed_uuid (ExmManager  *self,
     return FALSE;
 }
 
-static void
-do_install_thread (GTask        *task,
-                   ExmManager   *self,
-                   const char   *uuid,
-                   GCancellable *cancellable)
-{
-    GError *error = NULL;
-
-    g_dbus_proxy_call_sync (G_DBUS_PROXY (self->proxy),
-                            "InstallRemoteExtension",
-                            g_variant_new ("(s)", uuid, NULL),
-                            G_DBUS_CALL_FLAGS_NONE,
-                            -1, cancellable, &error);
-
-    if (error != NULL)
-    {
-        g_task_return_error (task, error);
-        return;
-    }
-
-    g_task_return_boolean (task, TRUE);
-}
-
 void
 exm_manager_install_async (ExmManager          *self,
                            const gchar         *uuid,
@@ -383,22 +360,35 @@ exm_manager_install_async (ExmManager          *self,
                            GAsyncReadyCallback  callback,
                            gpointer             user_data)
 {
-    GTask *task;
-
-    task = g_task_new (self, cancellable, callback, user_data);
-    g_task_set_task_data (task, g_strdup (uuid), (GDestroyNotify) g_free);
-    g_task_run_in_thread (task, (GTaskThreadFunc)do_install_thread);
-    g_object_unref (task);
+    shell_extensions_call_install_remote_extension (self->proxy,
+                                                    uuid,
+                                                    cancellable,
+                                                    callback,
+                                                    user_data);
 }
 
 gboolean
-exm_manager_install_finish (ExmManager    *self,
+exm_manager_install_finish (GObject       *self,
                             GAsyncResult  *result,
                             GError       **error)
 {
-    g_return_val_if_fail (g_task_is_valid (result, self), FALSE);
+    gchar *out_result = NULL;
+    gboolean success;
 
-    return g_task_propagate_boolean (G_TASK (result), error);
+    g_return_val_if_fail (SHELL_IS_EXTENSIONS (self), FALSE);
+
+    success = shell_extensions_call_install_remote_extension_finish (SHELL_EXTENSIONS (self),
+                                                                     &out_result,
+                                                                     result,
+                                                                     error);
+
+    if (g_str_equal (out_result, "cancelled"))
+    {
+        success = FALSE;
+        g_free (out_result);
+    }
+
+    return success;
 }
 
 static int
@@ -523,6 +513,14 @@ exm_manager_class_init (ExmManagerClass *klass)
                         0, NULL, NULL, NULL,
                         G_TYPE_NONE, 1,
                         G_TYPE_STRING);
+
+    signals [SIGNAL_INSTALL_STATUS]
+        = g_signal_new ("install-status",
+                        G_TYPE_FROM_CLASS (object_class),
+                        G_SIGNAL_RUN_LAST|G_SIGNAL_NO_RECURSE|G_SIGNAL_NO_HOOKS,
+                        0, NULL, NULL, NULL,
+                        G_TYPE_NONE, 1,
+                        EXM_TYPE_INSTALL_BUTTON_STATE);
 }
 
 static void
@@ -536,18 +534,21 @@ parse_single_extension (ExmExtension **extension,
 
     // Well-Defined Properties
     gchar *uuid = NULL;
-    gchar *display_name = NULL;
+    gchar *name = NULL;
     gchar *description = NULL;
+    ExmExtensionState state = EXM_EXTENSION_STATE_INITIALIZED;
     gboolean enabled = FALSE;
-    gboolean is_user = FALSE;
+    gchar *url = NULL;
+    gchar *version = NULL;
+    gchar *version_name = NULL;
+    gchar *error = NULL;
     gboolean has_prefs = FALSE;
     gboolean has_update = FALSE;
     gboolean can_change = TRUE;
-    ExmExtensionState state = EXM_EXTENSION_STATE_ACTIVE;
+    gboolean is_user = FALSE;
     ExmExtensionType type = EXM_EXTENSION_TYPE_SYSTEM;
-    gchar *version = NULL;
-    gchar *version_name = NULL;
-    gchar *error_msg = NULL;
+    GPtrArray *session_modes = g_ptr_array_new_with_free_func (g_free);
+    GHashTable *donations = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 
     if (extension && *extension)
     {
@@ -567,88 +568,121 @@ parse_single_extension (ExmExtension **extension,
         g_debug (" - Property: %s=%s\n", prop_name, g_variant_print(prop_value, 0));
 
         // Compare with DBus property names
-        if (strcmp (prop_name, "uuid") == 0)
+        if (g_strcmp0 (prop_name, "uuid") == 0)
         {
             g_variant_get (prop_value, "s", &uuid);
 
             // Assert that this is the same as the extension uuid
-            g_assert (strcmp(extension_uuid, uuid) == 0);
+            g_assert (g_strcmp0 (extension_uuid, uuid) == 0);
         }
-        else if (strcmp (prop_name, "type") == 0)
+        else if (g_strcmp0 (prop_name, "name") == 0)
         {
-            gdouble val;
-            g_variant_get (prop_value, "d", &val);
-            type = (ExmExtensionType)val;
+            g_variant_get (prop_value, "s", &name);
         }
-        else if (strcmp (prop_name, "state") == 0)
+        else if (g_strcmp0 (prop_name, "description") == 0)
+        {
+            g_variant_get (prop_value, "s", &description);
+        }
+        else if (g_strcmp0 (prop_name, "state") == 0)
         {
             gdouble val;
             g_variant_get (prop_value, "d", &val);
             state = (ExmExtensionState)val;
         }
-        else if (strcmp (prop_name, "enabled") == 0)
+        else if (g_strcmp0 (prop_name, "enabled") == 0)
         {
             g_variant_get (prop_value, "b", &enabled);
         }
-        else if (strcmp (prop_name, "name") == 0)
+        else if (g_strcmp0 (prop_name, "url") == 0)
         {
-            g_variant_get (prop_value, "s", &display_name);
+            g_variant_get (prop_value, "s", &url);
         }
-        else if (strcmp (prop_name, "description") == 0)
-        {
-            g_variant_get (prop_value, "s", &description);
-        }
-        else if (strcmp (prop_name, "hasPrefs") == 0)
-        {
-            g_variant_get (prop_value, "b", &has_prefs);
-        }
-        else if (strcmp (prop_name, "hasUpdate") == 0)
-        {
-            g_variant_get (prop_value, "b", &has_update);
-        }
-        else if (strcmp (prop_name, "canChange") == 0)
-        {
-            g_variant_get (prop_value, "b", &can_change);
-        }
-        else if (strcmp (prop_name, "version") == 0)
+        else if (g_strcmp0 (prop_name, "version") == 0)
         {
             gdouble val;
             g_variant_get (prop_value, "d", &val);
             version = g_strdup_printf ("%d", (gint)val);
         }
-        else if (strcmp (prop_name, "version-name") == 0)
+        else if (g_strcmp0 (prop_name, "version-name") == 0)
         {
             g_variant_get (prop_value, "s", &version_name);
         }
-        else if (strcmp (prop_name, "error") == 0)
+        else if (g_strcmp0 (prop_name, "error") == 0)
         {
-            g_variant_get (prop_value, "s", &error_msg);
+            g_variant_get (prop_value, "s", &error);
+        }
+        else if (g_strcmp0 (prop_name, "hasPrefs") == 0)
+        {
+            g_variant_get (prop_value, "b", &has_prefs);
+        }
+        else if (g_strcmp0 (prop_name, "hasUpdate") == 0)
+        {
+            g_variant_get (prop_value, "b", &has_update);
+        }
+        else if (g_strcmp0 (prop_name, "canChange") == 0)
+        {
+            g_variant_get (prop_value, "b", &can_change);
+        }
+        else if (g_strcmp0 (prop_name, "type") == 0)
+        {
+            gdouble val;
+            g_variant_get (prop_value, "d", &val);
+            type = (ExmExtensionType)val;
+        }
+        else if (g_strcmp0 (prop_name, "sessionModes") == 0)
+        {
+            GVariantIter iter;
+            GVariant *variant;
+            gchar *mode;
+            g_variant_iter_init (&iter, prop_value);
+            while (g_variant_iter_loop (&iter, "v", &variant))
+            {
+                mode = g_variant_dup_string (variant, NULL);
+                g_ptr_array_add (session_modes, g_strdup (mode));
+            }
+        }
+        else if (g_strcmp0 (prop_name, "donations") == 0)
+        {
+            GVariantIter iter;
+            gchar *key;
+            g_variant_iter_init (&iter, prop_value);
+            while (g_variant_iter_loop (&iter, "{sv}", &key, &prop_value))
+            {
+                const gchar *donation_value = g_variant_get_string (prop_value, NULL);
+                g_hash_table_insert (donations, g_strdup (key), g_strdup (donation_value));
+            }
         }
     }
 
+    // TODO: Show only version name when available in GNOME Extensions API
+    if (version_name)
+        version = g_strdup_printf ("%s (%s)", version, version_name);
     is_user = (type == EXM_EXTENSION_TYPE_PER_USER);
     *is_uninstall_operation = (state == EXM_EXTENSION_STATE_UNINSTALLED);
 
     g_object_set (*extension,
-                  "display-name", display_name,
+                  "name", name,
                   "description", description,
                   "state", state,
                   "enabled", enabled,
-                  "is-user", is_user,
+                  "url", url,
+                  "version", version,
+                  "error", error,
                   "has-prefs", has_prefs,
                   "has-update", has_update,
                   "can-change", can_change,
-                  "version", version,
-                  "version_name", version_name,
-                  "error-msg", error_msg,
+                  "is-user", is_user,
+                  "session-modes", session_modes,
+                  "donations", donations,
                   NULL);
 
     g_free (uuid);
-    g_free (display_name);
+    g_free (name);
     g_free (description);
+    g_free (url);
     g_free (version);
     g_free (version_name);
-    g_free (error_msg);
+    g_free (error);
 }
 
 static void
@@ -712,16 +746,6 @@ update_extension_list (ExmManager *self)
     queue_notify_extension_updates (self);
 }
 
-static gboolean
-is_extension_equal (ExmExtension *a, ExmExtension *b)
-{
-    const gchar *uuid_a, *uuid_b;
-    g_object_get (a, "uuid", &uuid_a, NULL);
-    g_object_get (b, "uuid", &uuid_b, NULL);
-
-    return strcmp (uuid_a, uuid_b) == 0;
-}
-
 static void
 on_state_changed (ShellExtensions *object G_GNUC_UNUSED,
                   const gchar     *arg_uuid,
@@ -752,13 +776,14 @@ on_state_changed (ShellExtensions *object G_GNUC_UNUSED,
 
     if (is_new)
     {
-        g_list_store_append (list_store, extension);
+        g_list_store_insert_sorted (list_store, extension, (GCompareDataFunc)compare_extension, NULL);
         return;
     }
 
+    guint position;
+
     if (is_uninstall_operation)
     {
-        guint position;
         if (g_list_store_find_with_equal_func (list_store, extension, (GEqualFunc)is_extension_equal, &position))
             g_list_store_remove (list_store, position);
 
@@ -766,11 +791,8 @@ on_state_changed (ShellExtensions *object G_GNUC_UNUSED,
     }
 
     // Emit items-changed signal to re-sort extension list
-    {
-        guint position;
-        if (g_list_store_find_with_equal_func (list_store, extension, (GEqualFunc)is_extension_equal, &position))
-            g_list_model_items_changed (G_LIST_MODEL (list_store), position, 1, 1);
-    }
+    if (g_list_store_find_with_equal_func (list_store, extension, (GEqualFunc)is_extension_equal, &position))
+        g_list_model_items_changed (G_LIST_MODEL (list_store), position, 0, 0);
 
     // If the extension that has changed has an update, then
     // one or more extensions have updates available. Lazily
